export type EventHandler<T> = (payload: T) => void;

export type ListenerRef = {
	off(): void;
};

export type EventHandlerMap<TEventMap, T extends keyof TEventMap> = Record<
	T,
	EventHandler<TEventMap[T]>[]
>;

export class EventEmitter<TEventMap> {
	private readonly listeners: EventHandlerMap<TEventMap, keyof TEventMap> =
		{} as EventHandlerMap<TEventMap, keyof TEventMap>;

	public on<T extends keyof TEventMap>(
		type: T,
		handler: EventHandler<TEventMap[T]>
	): ListenerRef {
		if (!this.listeners[type]) {
			this.listeners[type] = [];
		}
		this.listeners[type].push(handler);

		return {
			off() {
				this.off(type, handler);
			},
		};
	}

	public off<T extends keyof TEventMap>(
		type: T,
		handler: EventHandler<TEventMap[T]>
	) {
		if (this.listeners[type]) {
			this.listeners[type] = this.listeners[type].filter(
				(h) => h !== handler
			);
		}
	}

	public once<T extends keyof TEventMap>(
		type: T,
		handler: EventHandler<TEventMap[T]>
	) {
		const wrappedHandler = (payload: TEventMap[T]) => {
			handler(payload);
			this.off(type, wrappedHandler);
		};
		this.on(type, wrappedHandler);
	}

	public emit<T extends keyof TEventMap>(type: T, payload: TEventMap[T]) {
		if (this.listeners[type]) {
			this.listeners[type].forEach((handler) => handler(payload));
		}
	}
}
